(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_sched.ml 314 2010-10-31 12:24:56Z gerd $ *)

(** A plan is a directed graph. We are building the graph here iteratively -
   not all information are available at the beginning.

   The vertices are either:
    - files 
    - tasks

   Edges go from files to tasks (i.e. the files are inputs of the task), 
   and from tasks to files (i.e. the files are outputs of the task).

   The files may get tags:
   - [`Input]: The file is an input file for the whole workflow
   - [`Map_out]: The file is the result of a map job
   - [`Sort_out]: The file is the result of a sort job
   - [`Shuffle_out]: The file is the result of a shuffle job
   - [`Output]: The file is an output file for the whole workflow

   We provide an index mapping tags to vertices.

   Also, we maintain a few sets while constructing the graph:
   - [input_rest]: [`Input] files that are not yet consumed by a task
     (i.e. there is no edge going out of the file vertex)
   - [map_out_rest]: not yet consumed [`Map_out] files
   - [sort_out_rest]: not yet consumed [`Sort_out] files
   - [shuffle_out_rest(round,pmin,pmax)]: not yet consumed [`Shuffle_out] files
     for this [round] of shuffling and the input partitions [pmin] to [pmax]

   During the process of graph construction, vertexes may be added to or
   removed from these sets. Additions happen when more inputs become
   available, and removals when files get connected to tasks processing
   them.

   There also booleans: [input_complete], [map_out_complete], etc which
   are true when it is known that no more additions can occur.

 *)

open Mapred_tasks
open Plasma_rpcapi_aux
open Printf


type tag_type =
    [ `Input
    | `Map_out
    | `Sort_out
    | `Shuffle_out of int * int * int  (* round, pmin, pmax *)
    | `Output
    ]

type tag =
    [ `Input of int
	(* arg: the ordinal number of the input file, 0..n-1 *)
    | `Map_out of int
	(* arg: the ordinal number of the map output file, 0..m-1 *)
    | `Sort_out of int
	(* arg: the ordinal number of the sort output file, 0..m-1 *)
    | `Shuffle_out of int * int * int * int * int
	(* arg (round, part_min, part_max, file_min, file_max) *)
    | `Output of int
	(* arg: partition *)
    ]


let type_of_tag =
  function
    | `Input _ -> `Input
    | `Map_out _ -> `Map_out
    | `Sort_out _ -> `Sort_out
    | `Shuffle_out(round,pmin,pmax,_,_) -> `Shuffle_out(round,pmin,pmax)
    | `Output _ -> `Output

module Tag = struct
  type t = tag

  let rank_of_tt =
    function
      | `Input -> 0
      | `Map_out -> 1
      | `Sort_out -> 2
      | `Shuffle_out _ -> 3
      | `Output -> 4

  let ( ||| ) x y =
    if x = 0 then y else x

  let compare_tt xt yt =
    match xt, yt with
      | `Shuffle_out(xround,xpmin,xpmax), `Shuffle_out(yround,ypmin,ypmax) -> 
	  (xround - yround) ||| (xpmin - ypmin) ||| (xpmax - ypmax)
      | _ -> rank_of_tt xt - rank_of_tt yt

  let compare (x:t) (y:t) =
    match x, y with
      | `Input p, `Input q -> p - q
      | `Map_out p, `Map_out q -> p - q
      | `Sort_out p, `Sort_out q -> p - q
      | `Shuffle_out (xround, xpmin, xpmax, xfmin, xfmax), 
	  `Shuffle_out (yround, ypmin, ypmax, yfmin, yfmax) ->
	  (xround - yround) ||| (xpmin - ypmin) ||| (xpmax - ypmax) |||
	      (xfmin - yfmin) ||| (xfmax - yfmax)
      | `Output p, `Output q -> p - q
      | _ ->
	  let xt = type_of_tag x in
	  let yt = type_of_tag y in
	  rank_of_tt xt - rank_of_tt yt
    
end


module TagSet = Set.Make(Tag)

module V = struct
  type t =
      [ `File of file
      | `Task of task
      ]
  let compare (x:t) (y:t) = Pervasives.compare x y
  let equal (x:t) (y:t) = (x=y)
  let hash (x:t) = Hashtbl.hash x
end


(** Graph module: *)
module G = Graph.Imperative.Digraph.ConcreteBidirectional(V)

module Task = struct
  type t = task
  let compare (x:t) (y:t) = Pervasives.compare x y
end

module TaskSet = Set.Make(Task)


type plan_config =
    { keep_temp_files : bool;
      task_servers : string list;
      task_servers_ip : Unix.inet_addr list;
      jc : Mapred_def.mapred_job_config;
      mr_config : Mapred_config.mapred_config;
    }


type plan =
    { config : plan_config;
      cluster : Plasma_client.plasma_cluster;
      bigblock_size : int;  (* the real one *)
      graph : G.t;
      tagidx : (tag, V.t list) Hashtbl.t;
        (** Maps tag to the vertexes *)
      vtxtag : (V.t, tag) Hashtbl.t;
        (** Maps vertex to tag *)
      set : (tag_type, TagSet.t) Hashtbl.t;
        (** The mentioned sets per [tag_type] *)
      complete : (tag_type, bool) Hashtbl.t;
        (** Whether the sets are complete *)
      max : (tag_type, tag) Hashtbl.t;
        (** The maximum tag per tag type, but only for consumed files *)
      mutable plan_complete : bool;
        (** Whether the plan is fully complete *)
      mutable next_input_id : int;
      mutable next_map_out_id : int;
      mutable next_shuffle_id : int;
      mutable next_reduce_id : int;
      mutable vertex_added : bool;
        (** Used to determine whether the graph has been extended *)
      mutable runnable : TaskSet.t;
      mutable running : TaskSet.t;
      mutable finished : TaskSet.t;
      mutable missing_output : TaskSet.t;
        (** Map tasks where we did not yet got all output files *)
    }

(**********************************************************************)

(** {2 Determining runnable tasks} *)

let mark_as_started plan task =
  plan.runnable <- TaskSet.remove task plan.runnable;
  plan.running <- TaskSet.add task plan.running;
  plan.finished <- TaskSet.remove task plan.finished

let debug_marking = true


let maybe_mark_as_runnable plan task =
  (* Netlog.logf `Info
    "maybe_mark_as_runnable task=%s" (string_of_task_id task);
   *)

  let l1 = G.pred plan.graph (`Task task) in
  let l2 = List.flatten (List.map (fun v -> G.pred plan.graph v) l1) in
  (* Netlog.logf `Info
    "maybe_mark_as_runnable #l2=%d" (List.length l2);
   *)
  let old_runnable = TaskSet.mem task plan.runnable in
  let runnable =
    List.for_all
      (fun v ->
	 match v with
	 | `Task t ->
	     TaskSet.mem t plan.finished
	 | `File _ -> assert false
      )
      l2 in
  (* Netlog.logf `Info
    "maybe_mark_as_runnable runnable=%B" runnable;
   *)

  if debug_marking then (
    Netlog.logf `Debug "  succ -> %s: was_runnable=%B new_runnable=%B"
      (Mapred_tasks.string_of_task_id task) old_runnable runnable;
    List.iter
      (function
	 | `Task t ->
	     Netlog.logf `Debug "    pred -> %s: finished=%B"
	       (Mapred_tasks.string_of_task_id t) 
	       (TaskSet.mem t plan.finished)
	 | `File _ -> assert false
      )
      l2
  );

  if runnable then
    plan.runnable <- TaskSet.add task plan.runnable
  

let mark_as_finished plan task =
  let again = TaskSet.mem task plan.finished in
  plan.runnable <- TaskSet.remove task plan.runnable;
  plan.running <- TaskSet.remove task plan.running;
  plan.finished <- TaskSet.add task plan.finished;
  if debug_marking then (
    Netlog.logf `Debug "Mapred_sched.mark_as_finished (again=%B)" again;
    Netlog.logf `Debug "%s:" (Mapred_tasks.string_of_task_id task);
    let is_reduce =
      match task with
	| `Shuffle t -> t.shuffle_reduce
	| _ -> false in
    if is_reduce then
      Netlog.logf `Debug "  is a reduce task"
  );
  try
    let l1 = G.succ plan.graph (`Task task) in
    let l2 = List.flatten (List.map (fun v -> G.succ plan.graph v) l1) in
    (*Netlog.logf `Info
      "mark_as_finished task=%s #l2=%d"
      (string_of_task_id task) (List.length l2); *)
    List.iter
      (fun v ->
	 match v with
	   | `Task t ->
	       maybe_mark_as_runnable plan t
		 (* t is runnable if all predecessors are finished *)
	   | `File _ -> ()
      )
      l2
  with
    | Invalid_argument _ -> ()
    

let maybe_mark_as_finished_again plan task =
  (** If successors of [task] are later added, and [task] is already finished,
     we have to check again whether the successors are runnable.
   *)
  if TaskSet.mem task plan.finished then
    mark_as_finished plan task


let check_marks_for_new_task plan task =
  (** After adding [task], check whether there are predecessors that are
      now runnable
   *)
  let l1 = G.pred plan.graph (`Task task) in
  let l2 = List.flatten (List.map (fun v -> G.pred plan.graph v) l1) in
  List.iter
    (function
       | `Task t ->
	   maybe_mark_as_finished_again plan t
       | `File _ -> assert false
    )
    l2


let plan_finished plan =
  plan.plan_complete &&
  plan.runnable = TaskSet.empty && 
  plan.running = TaskSet.empty && 
  plan.finished <> TaskSet.empty

let n_running plan =
  TaskSet.cardinal plan.running

let n_finished plan =
  TaskSet.cardinal plan.finished

let n_total plan =
  G.fold_vertex
    (fun v acc ->
       match v with
	 | `Task _ -> acc+1
	 | _ -> acc
    )
    plan.graph
    0


(**********************************************************************)

(** {2 Graph construction in general} *)

let get_set plan tt =
  try Hashtbl.find plan.set tt with Not_found -> TagSet.empty

let get_complete plan tt =
  try Hashtbl.find plan.complete tt with Not_found -> false

let find_vertex plan tag =
  (** Find a _single_ vertex by tag *)
  try
    match Hashtbl.find plan.tagidx tag with
      | [v] -> v
      | _ -> raise Not_found 
  with
    | Not_found ->
	assert false

let find_file_vertex plan tag =
  (** Find a single file vertex by tag *)
  match find_vertex plan tag with
    | `File file -> file
    | _ -> assert false
  


let add_task plan task =
  let v = `Task task in
  G.add_vertex plan.graph v;
  plan.vertex_added <- true;
  v

let add_file plan (f:file) (t:tag) =
  let v = (`File f) in
  G.add_vertex plan.graph v;
  Hashtbl.replace plan.vtxtag v t;
  Hashtbl.replace plan.tagidx t [v];
    (* right now the tags are unique per file *)
  let tt = type_of_tag t in
  let s = get_set plan tt in
  let s' = TagSet.add t s in
  Hashtbl.replace plan.set tt s';
  plan.vertex_added <- true;
  v

let add_edge plan (v1:V.t) (v2:V.t) =
  (** Add the edge, and update [set] *)
  G.add_edge plan.graph v1 v2;
  ( try
      let t = Hashtbl.find plan.vtxtag v1 in
      let tt = type_of_tag t in
      let s = get_set plan tt in
      let s' = TagSet.remove t s in
      Hashtbl.replace plan.set tt s';
      ( try
	  let m = Hashtbl.find plan.max tt in
	  if Tag.compare t m > 0 then raise Not_found
	with
	  | Not_found ->
	      Hashtbl.replace plan.max tt t
      );
    with
      | Not_found -> ()
  )


module DFS = Graph.Traverse.Dfs(G)

let string_of_tag_type =
  function
    | `Input -> "Input"
    | `Map_out -> "Map_out"
    | `Sort_out -> "Sort_out"
    | `Shuffle_out(round,pmin,pmax) ->
	sprintf "Shuffle_out(%d,%d,%d)" round pmin pmax
    | `Output -> "Output"

let string_of_tag =
  function
    | `Input n -> sprintf "Input(%d)" n
    | `Map_out(n) -> sprintf "Map_out(%d)" n
    | `Sort_out n -> sprintf "Sort_out(%d)" n
    | `Shuffle_out(round,pmin,pmax,fmin,fmax) ->
	sprintf "Shuffle_out(%d,%d,%d,%d,%d)" round pmin pmax fmin fmax
    | `Output p -> sprintf "Output(%d)" p


let print_plan plan =
  DFS.prefix
    (fun v ->
       match v with
	 | `Task t ->
	     printf "Task:\n";
	     Mapred_tasks.print_task t 4;
	     printf "    Inputs:\n";
	     List.iter
	       (fun vp ->
		  match vp with
		    | `Task _ -> assert false
		    | `File f ->
			print_file f 8
	       )
	       (G.pred plan.graph v);
	     printf "    Outputs:\n";
	     List.iter
	       (fun vp ->
		  match vp with
		    | `Task _ -> assert false
		    | `File f ->
			print_file f 8
	       )
	       (G.succ plan.graph v)
	     
	 | `File _ -> ()
    )
    plan.graph;

  Hashtbl.iter
    (fun t l ->
       printf "Tag %s:\n" (string_of_tag t);
       List.iter
	 (fun vp ->
	    match vp with
	      | `Task _ -> ()
	      | `File f ->
		  print_file f 8
	 )
	 l
    )
    plan.tagidx;

  Hashtbl.iter
    (fun tt set ->
       printf "Tag type %s: set size = %d\n"
	 (string_of_tag_type tt)
	 (TagSet.cardinal set)
    )
    plan.set;

  Hashtbl.iter
    (fun tt flag ->
       printf "Tag type %s: complete = %B\n"
	 (string_of_tag_type tt)
	 flag
    )
    plan.complete


(**********************************************************************)

(** {2 Adding [`Input] files to the plan} *)


let ranked_hosts servertab blocks =
  (** Get the list of hosts storing [blocks] as list of pairs [(p,node)]
      where [p] is the share (range 0 to 1), and [node] is the IP address
   *)
  let ht = Hashtbl.create 5 in
  let n = ref 0 in
  List.iter
    (fun b ->
       if b.node_alive then (
	 let (h,p) = Plasma_util.parse_host_port b.node in
	 let ip =
	   try Unix.inet_addr_of_string h 
	   with _ ->
	     failwith ("Mapred_sched.ranked_hosts: not an IP address: " ^ h) in
	 if Hashtbl.mem servertab ip then (
	   let k = try Hashtbl.find ht ip with Not_found -> 0 in
	   Hashtbl.replace ht ip (k+1);
	   incr n
	 )
       )
    )
    blocks;
  let l = Hashtbl.fold (fun ip k acc -> (ip,k)::acc) ht [] in
  let l' = List.map (fun (ip,k) -> (float k /. float !n, ip)) l in
  List.sort
    (fun (p1,_) (p2,_) -> compare p1 p2)
    l'


let add_inputs plan =
  (** For every host we determine the number of map tasks it should get. 

      We iterate over the files and the blocklists. Look for sequences
      that are on the same host. When we have identified such a sequence,
      add it to the map task list of this host. If this list is already
      full, take another host randomly.
   *)
  let pc = plan.config in
  let c = plan.cluster in

  let n_hosts = List.length pc.task_servers in

  let servertab = Hashtbl.create 5 in   (** all task servers *)
  List.iter
    (fun ip -> Hashtbl.replace servertab ip ())
    pc.task_servers_ip;

  let trans = Plasma_client.start c in
  let blocksize = Plasma_client.blocksize c in
  let files = Plasma_client.list trans plan.config.jc#input_dir in
  let files_ii =
    List.map
      (fun (file,inode) -> 
	 (* Ignore inode, and do a lookup on file. This will follow
	    symlinks
	  *)
	 let abs_file = plan.config.jc#input_dir ^ "/" ^ file in
	 let real_inode = Plasma_client.lookup trans abs_file false in
	 let ii = Plasma_client.get_inodeinfo trans real_inode in
	 (file,real_inode,ii)
      )
      files in
  let total_blocksL =
    List.fold_left
      (fun acc (_,_,ii) ->
	 Int64.add acc ii.blocklimit
      )
      0L
      files_ii in

  let map_tasks =
    if pc.jc#map_tasks = 0 then (
      let mt_for_max_load =
	truncate
	  (float
	     (List.length pc.task_servers) *. 
	     pc.mr_config#mr_task_load_limit *. 4.0) in
      let bb_multiple =
	plan.bigblock_size / blocksize in
      let mt_minimum =
	max
	  (Int64.to_int
	     (Int64.div total_blocksL (Int64.of_int bb_multiple)))
	  1 in
      min mt_minimum mt_for_max_load
    )
    else
      pc.jc#map_tasks in

  Netlog.logf `Debug
    "Mapred_sched.add_inputs: map_tasks=%d" map_tasks;

  let n_tasks_per_host = map_tasks / n_hosts + 1 in
  let n_blocks_per_taskL = 
    Int64.div total_blocksL (Int64.of_int map_tasks) in

  let tasks_per_host = Hashtbl.create 5 in
  (** number of tasks assigned to each host *)

  let cur_task = Hashtbl.create 5 in
  (** this is the current task: [(size, list)]. [size] in blocks.
      [list] is a list of tuples [(filename,pos,len)].
   *)

  let add_map_task ip (f:file) =
    (** Adds the input file vertex for [f], plus the map task vertex.
	There is an edge from the input file to the task.
     *)
    let f =
      List.map
	(fun (filename,pos,len) ->
	   sprintf "%s/%s" plan.config.jc#input_dir filename, pos, len
	)
	f in
    let id = plan.next_input_id in
    plan.next_input_id <- id+1;
    let v_in_file = add_file plan f (`Input id) in
    let task =
      `Map
	{ map_input = f;
	  map_output_prefix = 
	    sprintf "%s/mapped_%06d_" plan.config.jc#work_dir id;
	  map_id = id;
	  map_best_hosts = [ip];
	} in
    let v_task = add_task plan task in
    plan.runnable <- TaskSet.add task plan.runnable;
    plan.missing_output <- TaskSet.add task plan.missing_output;
    add_edge plan v_in_file v_task;
    (** The output files for this map task cannot yet be added, as they
	are not yet known
     *)
  in

  (** analyze in groups of bigblocks *)
  let sched_limit = plan.bigblock_size in
  let sched_size = sched_limit / blocksize in
  let sched_sizeL = Int64.of_int sched_size in

  List.iter
    (fun (filename,inode,ii) ->
       let n_groupsL = 
	 Int64.succ (Int64.div (Int64.pred ii.blocklimit) sched_sizeL) in
       if n_groupsL > Int64.of_int max_int then
	 failwith "Mapred_sched: input file has too many blocks";
       (* well, that limit is really high, on 32 bit systems:
	  max_int = 1G - 1, so the max size is roughly 1G * blocksize,
	  e.g. 1P for blocksize=1M. Also, you run out of memory way
	  before that!
	*)
       
       let n_groups = Int64.to_int n_groupsL in
       for g = 0 to n_groups - 1 do
	 (** Retrieve the blocklist (EXPENSIVE) *)
	 let pos0 = Int64.mul (Int64.of_int g) sched_sizeL in
	 let pos1 = min (Int64.add pos0 sched_sizeL) ii.blocklimit in
	 let lenL = Int64.sub pos1 pos0 in
	 let len = Int64.to_int lenL in
	 let blocks = Plasma_client.get_blocklist trans inode pos0 len in
	 (** Get a ranking of the nodes: *)
	 let nodes0 = ranked_hosts servertab blocks in
	 (** nodes0 may be empty for two reasons:
	    - all nodes dead
	    - a hole in the file
	    We check here only for the first. Holes will also be rejected
	    by the record reader, and it is much simpler to find out there.
	  *)
	 if nodes0 = [] then (
	   if blocks <> [] then
	     failwith ("File inaccessible (no live data nodes): " ^
			 plan.config.jc#input_dir ^ "/" ^ filename)
	 )
	 else (
	   let nodes = 
	     List.map (fun (p,ip) -> (p,ip,false)) nodes0 @
	       [ let (p,ip) = List.hd nodes0 in (p,ip,true) ] in
	   (** Try to assign the task: *)
	   ( try
	       (** iterate over nodes. Skip nodes when tasks_per_host is
		  >= n_tasks_per_hosts. Otherwise add the task to cur_task.
		  If cur_task is >= blocks_per_task, create the graph node.
		*)
	       List.iter
		 (fun (p,ip,is_last) ->
		    let n = 
		      try Hashtbl.find tasks_per_host ip 
		      with Not_found -> 0 in
		    if n < n_tasks_per_host || is_last then (
		      let ct_n, ct_l =
			try Hashtbl.find cur_task ip 
			with Not_found -> (0L, []) in
		      let ct_n' = Int64.add ct_n lenL in
		      let ct_l' = (filename,pos0,lenL) :: ct_l in
		      if ct_n' >= n_blocks_per_taskL then (
			Hashtbl.remove cur_task ip;
			add_map_task ip (List.rev ct_l');
			Hashtbl.replace tasks_per_host ip (n+1)
		      )
		      else
			Hashtbl.replace cur_task ip (ct_n', ct_l');
		      raise Exit
		    )
		 )
		 nodes
	     with
	       | Exit -> ()
	   )
	 );
       done
       
    )
    files_ii;

  Plasma_client.commit trans;

  (** Create the remaining tasks from [cur_task]: *)
  Hashtbl.iter
    (fun ip (ct_n, ct_l) ->
       add_map_task ip (List.rev ct_l);
    )
    cur_task


(**********************************************************************)

(** {2 Adding sort tasks} *)

let try_add_sort_tasks plan =
  (** If there is a [`Map_out] tag in [set], we can create a sort 
      task for it. Also check whether we have all sort tasks.
   *)
  let s = get_set plan `Map_out in
  let is_complete = get_complete plan `Map_out in
 
  TagSet.iter
    (fun tag ->
       let id =
	 match tag with
	   | `Map_out id -> id
	   | _ -> assert false in
       
       let map_out = find_file_vertex plan tag in
       let v_map_out = `File map_out in

       let sort_out_name =
	 sprintf "%s/sorted_%06d" plan.config.jc#work_dir id in
       let st =
	 { sort_input = map_out;
	   sort_input_del = not plan.config.keep_temp_files;
	   sort_output = sort_out_name;
	   sort_id = id
	 } in
       let sort_out = [ sort_out_name, 0L, 0L ] in
       let v_sort = add_task plan (`Sort st) in
       let v_sort_out = add_file plan sort_out (`Sort_out id) in
       add_edge plan v_map_out v_sort;
       add_edge plan v_sort v_sort_out;
       check_marks_for_new_task plan (`Sort st)
    )
    s;

  if is_complete then
    Hashtbl.replace plan.complete `Sort_out true

(**********************************************************************)

(** {2 Adding shuffle/reduce tasks} *)


let split_range n pmin pmax =
  (** splits the range of numbers pmin...pmax into n groups
     (g1min, g1max), (g2min, g2max), ..., (gnmin,gnmax)
     so that g(k)max = g(k+1)min, and the groups have approx the same size
     e.g.
     # split_range 4 0 10;;
     - : (int * int) list = [(0, 2); (3, 5); (6, 8); (9, 10)]
   *)
  let r = pmax-pmin+1 in   (* number of elements in total *)
  let n = min n r in
  let gsize = r/n in
  let gmod = r mod n in
  let l = ref [] in
  let p = ref pmin in
  for k = 0 to n-1 do
    let size = gsize + (if k < gmod then 1 else 0) in
    l := (!p, !p+size-1) :: !l;
    p := !p + size;
  done;
  List.rev !l


let new_shuffle_out_filename plan =
  let id = plan.next_shuffle_id in
  plan.next_shuffle_id <- id+1;
  sprintf "%s/shuffle_%09d" plan.config.jc#work_dir id


let log n =
  let rec mul k p =
    if p >= n then
      k
    else
      mul (k+1) (p*10) in
  mul 0 1


let new_reduce_out_filename plan =
  let id = plan.next_reduce_id in
  plan.next_reduce_id <- id+1;
  let width = log plan.config.jc#partitions in
  sprintf "%s/reduce_%0*d" plan.config.jc#output_dir width id


let try_add_first_round_shuffle_tasks plan =
  (** Shuffle tasks of the first round take the output of sort tasks.

      We can do something here when we have the next [n_merge] sort output
      files.
   *)

  let n_merge = plan.config.jc # merge_limit in
  let n_split = plan.config.jc # split_limit in

  let s = get_set plan `Sort_out in
  let is_complete = get_complete plan `Sort_out in

  let pmin = 0 in
  let pmax = plan.config.jc # partitions - 1 in

  let output_parts = split_range n_split pmin pmax in

  let id_of_sort_tag =
    function
      | `Sort_out id -> id
      | _ -> assert false in

  let rec collect id id_max =
    if id > id_max then
      []
    else
      let tag = (`Sort_out id) in
      if TagSet.mem tag s then
	tag :: collect (id+1) id_max
      else
	if is_complete then 
	  []
	else
	  raise Not_found
  in

  if s <> TagSet.empty then (
    let tag0 = TagSet.min_elt s in
    let id0 = id_of_sort_tag tag0 in
    (** We can only go on if [tag0] is really the first tag we need, i.e.
	the immediate predecessor (if any) is already processed.
     *)
    if id0=0 || Hashtbl.mem plan.tagidx (`Sort_out (id0-1)) then (
      (** Now check whether we even have the first [n_merge] tags *)
      try
	let tags = collect id0 (id0 + n_merge - 1) in
	  (** [collect] may raise [Not_found] *)
	let id_max = id_of_sort_tag (List.hd (List.rev tags)) in
	
	let shuffle_input =
	  List.map
	    (fun tag ->
	       let id = id_of_sort_tag tag in
	       let sort_out = find_file_vertex plan tag in
	       (sort_out, id, id)
	    )
	    tags in

	let shuffle_output =
	  List.map
	    (fun (qmin,qmax) ->
	       let name = new_shuffle_out_filename plan in
	       (name, qmin, qmax)
	    )
	    output_parts in
	
	let t =
	  { shuffle_input = shuffle_input;
	    shuffle_input_del = not plan.config.keep_temp_files;
	    shuffle_output = shuffle_output;
	    shuffle_partitions = (pmin,pmax);
	    shuffle_coverage = (id0,id_max);
	    shuffle_reduce = false;  (* never in the first round *)
	  } in

	let v_task = add_task plan (`Shuffle t) in
	List.iter
	  (fun tag ->
	     let v_sort_out = `File(find_file_vertex plan tag) in
	     add_edge plan v_sort_out v_task
	  )
	  tags;
	List.iter
	  (fun (name,qmin,qmax) ->
	     let file = [name, 0L, 0L] in
	     let tag = `Shuffle_out(0, qmin, qmax, id0, id_max) in
	     let v_shuffle_out = add_file plan file tag in
	     add_edge plan v_task v_shuffle_out
	  )
	  shuffle_output;
       check_marks_for_new_task plan (`Shuffle t)
      with Not_found -> ()      
    );

    let s' = get_set plan `Sort_out in
    if is_complete && s' = TagSet.empty then
      List.iter
	(fun (qmin,qmax) ->
	   Hashtbl.replace plan.complete (`Shuffle_out(0,qmin,qmax)) true
	)
	output_parts
  )

  
let try_add_these_shuffle_tasks plan round pmin pmax =
  (** Shuffle tasks of the following rounds take the output of previous
      shuffle tasks.

      round, pmin, pmax: these values identify the previous round!
   *)

  let n_merge = plan.config.jc # merge_limit in
  let n_split = plan.config.jc # split_limit in  

  let tt = `Shuffle_out(round,pmin,pmax) in
  let s = get_set plan tt in
  let is_complete = get_complete plan tt in

  let output_parts = split_range n_split pmin pmax in

  let ids_of_shuffle_tag =
    function
      | `Shuffle_out(_,_,_,fmin,fmax) -> (fmin,fmax)
      | _ -> assert false in

  let rec collect fmin_exp u n =
    if n = 0 then
      []
    else
      let tag_opt = try Some(TagSet.min_elt u) with Not_found -> None in
      match tag_opt with
	| Some tag ->
	    let (fmin_next, fmax_next) = ids_of_shuffle_tag tag in
	    if fmin_next = fmin_exp then
	      tag :: collect (fmax_next+1) (TagSet.remove tag u) (n-1)
	    else
	      raise Not_found
	| None ->
	    if is_complete then 
	      []
	    else
	      raise Not_found
  in

  (** Other than for the first round, we only start the next round if
      all input files are available. Otherwise we cannot reliably 
      determine whether we reach the end of the round.
   *)

  if s <> TagSet.empty && is_complete then (
    let tag0 = TagSet.min_elt s in
    let (fmin0,fmax0) = ids_of_shuffle_tag tag0 in
    (** We can only go on if [tag0] is really the first tag we need, i.e.
	the immediate predecessor (if any) is already processed.
     *)
    let pred_mem = (
      try
	let m = Hashtbl.find plan.max tt in
	let (fminm, fmaxm) = ids_of_shuffle_tag m in
	fmaxm+1 = fmin0
      with Not_found -> false
    ) in
    if fmin0=0 || pred_mem then (
      (** Now check whether we even have the first [n_merge] tags *)
      try
	let tags = collect fmin0 s n_merge in
	  (** [collect] may raise [Not_found] *)
	let (_,fmaxx) = ids_of_shuffle_tag (List.hd (List.rev tags)) in
	
	let shuffle_input =
	  List.map
	    (fun tag ->
	       let (fmin,fmax) = ids_of_shuffle_tag tag in
	       let file = find_file_vertex plan tag in
	       (file, fmin, fmax)
	    )
	    tags in

	let is_reduce = 
	  (pmin = pmax) && fmin0 = 0 && fmaxx+1 = plan.next_map_out_id in
	
	let shuffle_output =
	  if is_reduce then
	    [ new_reduce_out_filename plan, pmin, pmax ]
	  else
	    List.map
	      (fun (qmin,qmax) ->
		 let name = new_shuffle_out_filename plan in
		 (name, qmin, qmax)
	      )
	      output_parts in

	let t =
	  { shuffle_input = shuffle_input;
	    shuffle_input_del = not plan.config.keep_temp_files;
	    shuffle_output = shuffle_output;
	    shuffle_partitions = (pmin,pmax);
	    shuffle_coverage = (fmin0,fmaxx);
	    shuffle_reduce = is_reduce;
	  } in

	let v_task = add_task plan (`Shuffle t) in
	List.iter
	  (fun tag ->
	     let v_file = `File(find_file_vertex plan tag) in
	     add_edge plan v_file v_task
	  )
	  tags;
	List.iter
	  (fun (name,qmin,qmax) ->
	     let file = [name, 0L, 0L] in
	     let tag = 
	       if is_reduce then 
		 `Output qmin
	       else
		 `Shuffle_out(round+1, qmin, qmax, fmin0, fmaxx) in
	     let v_shuffle_out = add_file plan file tag in
	     add_edge plan v_task v_shuffle_out
	  )
	  shuffle_output;
       check_marks_for_new_task plan (`Shuffle t)
      with Not_found -> ()      
    );

    let s' = get_set plan tt in
    if s' = TagSet.empty then
      List.iter
	(fun (qmin,qmax) ->
	   Hashtbl.replace plan.complete (`Shuffle_out(round+1,qmin,qmax)) true
	)
	output_parts;

  )


let card_one s =
  (** quickly determine whether the cardinality of s is 1 *)
  let n = ref 0 in
  ( try
      TagSet.iter (fun e -> incr n; if !n > 1 then raise Exit) s
    with Exit -> ()
  );
  !n = 1


let try_add_shuffle_tasks plan =
  Hashtbl.iter
    (fun tt set ->
       match tt with
	 | `Shuffle_out(round,pmin,pmax) ->
	     try_add_these_shuffle_tasks plan round pmin pmax
	 | `Output ->
	     (* FIXME: [cardinal] is O(n) *)
	     if TagSet.cardinal set = plan.config.jc # partitions then
	       plan.plan_complete <- true
	 | _ ->
	     ()
    )
    plan.set

(**********************************************************************)

(** {2 Adding any task} *)

let add_tasks plan =
  let cont = ref true in
  while !cont do
    plan.vertex_added <- false;
    try_add_sort_tasks plan;
    try_add_first_round_shuffle_tasks plan;
    try_add_shuffle_tasks plan;
    cont := plan.vertex_added
  done;

  if get_complete plan `Map_out && not plan.plan_complete then (
    (* This is an error condition *)
    Netlog.logf `Err 
      "Plan is fully computed but not marked as complete! \
       (diagnostics to stdout)";
    printf "PLAN SO FAR:\n";
    print_plan plan;
    printf "next_reduce_id = %d\n" plan.next_reduce_id;
    printf "#missing_output = %d\n" (TaskSet.cardinal plan.missing_output);
    flush stdout
  )


let complete_map_outputs plan =
  Hashtbl.replace plan.complete `Map_out true


let complete_inputs plan =
  (** Mark the inputs as complete *)
  Hashtbl.replace plan.complete `Input true;
  if plan.missing_output = TaskSet.empty then (
    complete_map_outputs plan;
    add_tasks plan
  )


let add_map_output plan (mt:map_task) (files:file list) =
  let v_task = `Task(`Map mt) in
  if not (G.mem_vertex plan.graph v_task) then
    failwith "Mapred_sched.add_map_output";
  
  List.iter
    (fun file ->
       let id = plan.next_map_out_id in
       plan.next_map_out_id <- id+1;
       let t = `Map_out id in
       let v = add_file plan file t in
       add_edge plan v_task v
    )
    files;

  plan.missing_output <- TaskSet.remove (`Map mt) plan.missing_output;
  if plan.missing_output = TaskSet.empty then
    complete_map_outputs plan;

  add_tasks plan


(**********************************************************************)

(** {2 Interface} *)

let configure_plan ?(keep_temp_files=false) jc mr_config =
  if jc#bigblock_size <= 0 then
    invalid_arg "Mapred_sched.configure_plan: bad bigblock_size";
  let servers = mr_config # mr_task_nodes in
  let ips =
    List.map Plasma_util.ip_of_host servers in
  { keep_temp_files = keep_temp_files;
    task_servers = servers;
    task_servers_ip = ips;
    jc = jc;
    mr_config = mr_config;
  }


let create_plan c pc =
  let blocksize = Plasma_client.blocksize c in
  let bigblock_size = 
    ((pc.jc#bigblock_size - 1) / blocksize + 1) * blocksize in
  { config = pc;
    cluster = c;
    bigblock_size = bigblock_size;
    graph = G.create ~size:1000 ();
    tagidx = Hashtbl.create 1000;
    vtxtag = Hashtbl.create 50;
    set = Hashtbl.create 50;
    complete = Hashtbl.create 50;
    max = Hashtbl.create 50;
    plan_complete = false;
    next_input_id = 0;
    next_map_out_id = 0;
    next_shuffle_id = 0;
    next_reduce_id = 0;
    vertex_added = false;
    runnable = TaskSet.empty;
    running = TaskSet.empty;
    finished = TaskSet.empty;
    missing_output = TaskSet.empty;
  }

let bigblock_size plan =
  plan.bigblock_size

let executable_tasks plan =
  TaskSet.elements plan.runnable


let plan_complete plan =
  plan.plan_complete

let hosts plan =
  List.map2
    (fun n ip -> (n,ip))
    plan.config.task_servers
    plan.config.task_servers_ip

let task_depends_on_list plan task =
  let l =
    List.flatten
      (List.map (G.pred plan.graph) (G.pred plan.graph (`Task task))) in
  let l' =
    List.map
      (function
	 | `Task t -> t
	 | _ -> assert false
      )
      l in
  let s = List.fold_left (fun acc t -> TaskSet.add t acc) TaskSet.empty l' in
  TaskSet.elements s
  
let cluster plan =
  plan.cluster
