(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma.ml 285 2010-10-28 19:30:05Z gerd $ *)

open Printf
open Plasma_rpcapi_aux
open Plasma_util.Operators

let usage() =
  prerr_endline "usage: plasma <mode> [-help|options]";
  prerr_endline "";
  prerr_endline "  available modes:";
  prerr_endline "  - list | ls";
  prerr_endline "  - delete | rm";
  prerr_endline "  - rename | mv";
  prerr_endline "  - link | ln";
  prerr_endline "  - mkdir";
  prerr_endline "  - create";
  prerr_endline "  - put";
  prerr_endline "  - get";
  prerr_endline "  - cat";
  prerr_endline "  - fsstat";
  prerr_endline "  - blocks"


let access_cluster esys =
  let cluster = ref None in
  let namenodes = ref [] in
  let rep = ref 0 in
  let debug = ref false in
  let args =
    [ "-cluster", Arg.String (fun s -> cluster := Some s),
      "<name>    Set the cluster name";
      
      "-namenode", Arg.String (fun n -> namenodes := n :: !namenodes),
      "<host>:<port>   Also use this namenode (can be given several times)";

      "-rep", Arg.Set_int rep,
      "<n>   Set replication factor for new files (default 1)";

      "-debug", Arg.Set debug,
      "    Enable debug output";
    ] in
  ( object
      method args = args
      method cluster = !cluster
      method namenodes = !namenodes
      method rep = !rep
      method open_cluster() =
	Netlog.current_logger := 
	  Netlog.channel_logger stderr (if !debug then `Debug else `Info);
	let nn_nodes =
	  if !namenodes = [] then None else Some !namenodes in
	let cfg = Plasma_client_config.get_config
	            ?clustername:!cluster
                    ?nn_nodes () in
	let c = Plasma_client.open_cluster_cc cfg esys in
	c
    end
  )

let mtime_compare t1 t2 =
  let k = Int64.compare t1.tsecs t2.tsecs in
  if k = 0 then
    compare t1.tnsecs t2.tnsecs
  else
    k

let list() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  let one = ref false in
  let all = ref false in
  let dirs = ref false in
  let inode_flag = ref false in
  let reverse = ref false in
  let sort_by_size = ref false in
  let sort_by_mtime = ref false in
  Arg.parse
    ( ac#args @
	[ "-1", Arg.Set one,
	  "   output one filename per line and nothing else";

	  "-a", Arg.Set all,
	  "   also include entries starting with .";

	  "-d", Arg.Set dirs,
	  "   list directory entries instead of contents";

	  "-i", Arg.Set inode_flag,
	  "   output inode numbers";

	  "-r", Arg.Set reverse,
	  "   reverse sorting order";

	  "-S", Arg.Set sort_by_size,
	  "   sort by file size";

	  "-t", Arg.Set sort_by_mtime,
	  "   sort by modification time";
	] )
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma list <file> ...";

  let c = ac#open_cluster() in

  let show_rows rows =
    let widths =
      List.fold_left
	(fun acc row ->
	   if acc = [] then
	     List.map
	       (fun (field,_) -> String.length field)
	       row
	   else
	     List.map2
	       (fun width (field, alignment) ->
		  max width (String.length field)
	       )
	       acc row
	)
	[]
	rows in
    List.iter
      (fun row ->
	 let first = ref true in
	 List.iter2
	   (fun width (field, alignment) ->
	      if not !first then printf " ";
	      first := false;
	      match alignment with
		| `L -> printf "%-*s" width field
		| `R -> printf "%*s" width field
	   )
	   widths row;
	 printf "\n";
      )
      rows
  in    
	
  let file_row name inode ii =
    if !one then
      [ name, `L ]
    else (
      let rwx n =
	( if n land 4 <> 0 then "r" else "-" ) ^ 
	  ( if n land 2 <> 0 then "w" else "-" ) ^ 
	  ( if n land 1 <> 0 then "x" else "-" ) in
      let bits =
	( match ii.filetype with
	    | `ftype_regular -> "-"
	    | `ftype_directory -> "d"
	    | `ftype_symlink -> "s" 
	) ^
	  rwx (ii.mode lsr 6) ^ 
	  rwx (ii.mode lsr 3) ^ 
	  rwx ii.mode in
      let maybe_inode =
	if !inode_flag then [ (Int64.to_string inode), `R ] else [] in
      let mtime =
	(Netdate.format
	   ~fmt:"%Y-%m-%d %H:%M"
	   (Netdate.create 
	      ~zone:Netdate.localzone
	      (Int64.to_float ii.mtime.tsecs))) in
      let xname =
	if ii.filetype = `ftype_symlink then
	  name ^ " -> " ^ ii.field1
	else
	  name in
      let row =
	maybe_inode @
	  [ bits, `L;
	    ii.usergroup.user, `L;
	    ii.usergroup.group, `L;
	    Int64.to_string ii.eof, `R;
	    mtime, `R;
	    xname, `L
	  ] in
      row
    )
  in

  List.iter
    (fun file ->
       let t = Plasma_client.start c in
       let inode = Plasma_client.lookup t file false in
       let ii = Plasma_client.get_inodeinfo t inode in
       ( match ii.filetype with
	   | `ftype_directory when not !dirs ->
	       if not !one then printf "%s:\n%!" file;
	       let elements = 
		 List.flatten
		   (List.map
		      (fun (elem, elem_inode) ->
			 try
			   let elem_ii =
			     Plasma_client.get_inodeinfo t elem_inode in
			   [elem, elem_inode, elem_ii]
			 with
			   | Plasma_client.Plasma_error `enoent -> []
		      )
		      (Plasma_client.list_inode t inode)) in
	       let elements' =
		 List.filter
		   (fun (elem,_,_) ->
		      !all || (elem <> "" && elem.[0] <> '.')
		   ) elements in
	       let elements'' =
		 List.sort
		   (fun (elem1,_,elem_ii1) (elem2,_,elem_ii2) ->
		      let k =
			if !sort_by_size then
			  Int64.compare elem_ii1.eof elem_ii2.eof
			else
			  if !sort_by_mtime then
			    mtime_compare elem_ii1.mtime elem_ii2.mtime
			  else
			    String.compare elem1 elem2 in
		      if !reverse then -k else k
		   )
		   elements' in
	       let rows =
		 List.map
		   (fun (elem,elem_inode,elem_ii) ->
		      file_row elem elem_inode elem_ii
		   )
		   elements'' in
	       show_rows rows
	   | _ ->
	       show_rows [file_row file inode ii]
       );
       Plasma_client.commit t;
    )
    !files



let create() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma create <file> ...";

  let c = ac#open_cluster () in

  let ii = Plasma_client.regular_ii c 0o666 in
  let t = Plasma_client.start c in
  List.iter
    (fun name ->
       let inode = Plasma_client.create_inode t ii in
       eprintf "Created inode: %Ld\n%!" inode;
       Plasma_client.link t name inode;
       eprintf "Linked name: %s -> %Ld\n%!" name inode
    )
    !files;
  Plasma_client.commit t;
  eprintf "Committed.\n%!"


let mkdir() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma mkdir <file> ...";

  let c = ac#open_cluster () in

  let ii = Plasma_client.dir_ii c 0o777 in
  let t = Plasma_client.start c in
  List.iter
    (fun name ->
       let inode = Plasma_client.mkdir t name ii in
       eprintf "Created directory %s -> %Ld\n%!" name inode
    )
    !files;
  Plasma_client.commit t;
  eprintf "Committed.\n%!"


let link() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let dir = ref None in
  let symlink_flag = ref false in
  let ac = access_cluster esys in
  Arg.parse
    (ac#args @
       [ "-s", Arg.Set symlink_flag,
	 "    create symlink instead of hard link";

	 "-t", Arg.String (fun s -> dir := Some s),
	 "<dir>   create the links in this directory";
       ]
    )
    (fun s ->
       files := s :: !files
    )
    "usage: plasma ln <opts> (<target> <link> | <target> | <target>... <dir>) ";

  let c = ac#open_cluster () in
  let n = List.length !files in

  let t = Plasma_client.start c in

  if !dir = None && n = 1 then
    dir := Some ".";
  
  if !dir = None && n > 1 then (
    try
      let inode = Plasma_client.lookup t (List.hd !files) true in
      let ii = Plasma_client.get_inodeinfo t inode in
      if ii.filetype = `ftype_directory then (
	dir := Some (List.hd !files);
	files := List.tl !files
      )
    with
      | Plasma_client.Plasma_error (`enoent|`estale) -> ()
  );

  files := List.rev !files;

  let do_link target_name link_name =
    if !symlink_flag then (
      let ii = Plasma_client.symlink_ii c target_name in
      let _inode = Plasma_client.create_file t link_name ii in
      eprintf "Linked: %s -> %s\n%!" link_name target_name
    )
    else (
      let inode = Plasma_client.lookup t target_name true in
      Plasma_client.link t link_name inode;
      eprintf "Linked: %s -> %s\n%!" link_name target_name
    )
  in

  ( match !dir with
      | None ->
	  if n <> 2 then
	    failwith
	      "Exactly two arguments expected when no directory is specified";
	  ( match !files with
	      | [ target_name; link_name ] ->
		  do_link target_name link_name
	      | _ ->
		  assert false
	  )
	    
      | Some dirname ->
	  List.iter
	    (fun name ->
	       let basename = Filename.basename name in
	       do_link name (dirname ^ "/" ^ basename)
	    )
	    !files
  );
  Plasma_client.commit t;
  eprintf "Committed.\n%!"


let rename() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let dir = ref None in
  let ac = access_cluster esys in
  Arg.parse
    (ac#args @
       [ "-t", Arg.String (fun s -> dir := Some s),
	 "<dir>   move the files into this directory";
       ]
    )
    (fun s ->
       files := s :: !files
    )
    "usage: plasma ln <opts> (<src> <dest> | <src>... <destdir>) ";

  let c = ac#open_cluster () in
  let n = List.length !files in

  let t = Plasma_client.start c in

  if !dir = None && n > 1 then (
    try
      let inode = Plasma_client.lookup t (List.hd !files) true in
      let ii = Plasma_client.get_inodeinfo t inode in
      if ii.filetype = `ftype_directory then (
	dir := Some (List.hd !files);
	files := List.tl !files
      )
    with
      | Plasma_client.Plasma_error (`enoent|`estale) -> ()
  );

  files := List.rev !files;

  let do_rename src_name dest_name =
    Plasma_client.rename t src_name dest_name;
    eprintf "Renamed: %s -> %s\n%!" src_name dest_name
  in

  ( match !dir with
      | None ->
	  if n <> 2 then
	    failwith
	      "Exactly two arguments expected when no directory is specified";
	  ( match !files with
	      | [ src_name; dest_name ] ->
		  do_rename src_name dest_name
	      | _ ->
		  assert false
	  )
	    
      | Some dirname ->
	  List.iter
	    (fun name ->
	       let basename = Filename.basename name in
	     do_rename name (dirname ^ "/" ^ basename)
	    )
	    !files
  );
  Plasma_client.commit t;
  eprintf "Committed.\n%!"


let delete() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let recurse = ref false in
  let ac = access_cluster esys in

  Arg.parse
    (ac#args @
       [ "-r", Arg.Set recurse,
	 "   recursively delete directories"
       ]
    )
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma delete <file> ...";

  let c = ac#open_cluster () in
  let t = Plasma_client.start c in

  let rec remove_dir_contents dir_name dir_inode =
    let entries = Plasma_client.list_inode t dir_inode in
    List.iter
      (fun (base,_) -> remove_entry (dir_name ^ "/" ^ base) dir_inode base)
      entries
    
  and remove_entry name dir_inode base =
    Plasma_client.namelock t dir_inode base;
    try
      Plasma_client.unlink_at t dir_inode base;
      eprintf "Deleted: %s\n%!" name
    with
      | Plasma_client.Plasma_error `enotempty when !recurse ->
	  let inode = Plasma_client.dir_lookup t dir_inode base true in
	  let ii = Plasma_client.get_inodeinfo t inode in
	  if ii.filetype = `ftype_directory then
	    remove_dir_contents name inode;
	  Plasma_client.unlink_at t dir_inode base;
	  eprintf "Deleted: %s\n%!" name
  in

  List.iter
    (fun name ->
       let dir = Filename.dirname name in
       let base = Filename.basename name in
       let dir_inode = Plasma_client.lookup t dir false in
       remove_entry name dir_inode base
    )
    !files;
  Plasma_client.commit t;
  eprintf "Committed.\n%!"


let put() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  let plasma_filename = ref "" in
  let plasma_set = ref false in
  let local_filename = ref "" in
  let local_set = ref false in
  let topo = ref `Star in
  let from_stdin = ref false in
  let force = ref false in

  Arg.parse
    ( ac#args
      @ [ "-chain", Arg.Unit (fun () -> topo := `Chain),
	  "   Use chain topology instead of star topology";

	  "-stdin", Arg.Set from_stdin,
	  "   Copy data from stdin. No <local_filename> in this case.";

	  "-f", Arg.Set force,
	  "   Force: Replace existing files";
	]
    )
    (fun s ->
       if not !local_set && not !from_stdin then (
	 local_filename := s;
	 local_set := true
       )
       else
	 if not !plasma_set then (
	   plasma_filename := s;
	   plasma_set := true
	 )
	 else
	   raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma put (<local_path> <plasma_path> | -stdin <plasma_path>)";

  if not !plasma_set then
    failwith "No plasma filename given";   
  if not !local_set && not !from_stdin then
    failwith "No local filename given";
  if !local_set && !from_stdin then
    failwith "With -stdin, a localfile filename must not be given";

  if !from_stdin then
    local_filename := "<stdin>";

  let fd, len = 
    if !from_stdin then
      Unix.stdin, 0L
    else (
      let fd = Unix.openfile !local_filename [Unix.O_RDONLY] 0 in
      let len = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
      ignore(Unix.LargeFile.lseek fd 0L Unix.SEEK_SET);
      (fd, len)
    ) in
  
  let c = ac#open_cluster () in
  let ii = Plasma_client.regular_ii c 0o666 in
  let t = Plasma_client.start c in
  if !force then (
    try
      Plasma_client.unlink t !plasma_filename
    with Plasma_client.Plasma_error `enoent -> ()
  );
  let inode = Plasma_client.create_file t !plasma_filename ii in
  Plasma_client.commit t;
  Plasma_client.copy_in c inode 0L fd len !topo;
  Unix.close fd;

  eprintf "Copied: %s -> %s\n%!" !local_filename !plasma_filename


let alt_copy_out esys c inode pos fd len par =
  (* par: number of parallel requests *)

  Plasma_client.configure_buffer c (2*par);

  let t = Plasma_client.start c in
  let blocksize = Plasma_client.blocksize c in
  let blocksizeL = Int64.of_int blocksize in

  let stream_e, signal =
    Plasma_util.signal_engine esys in

  let pool = Netsys_mem.create_pool blocksize in

  let n = ref 0 in
  let p = ref pos in
  let end_pos = Int64.add pos len in

  let rec read_block_e p0 buf r bsize =
    Plasma_client.read_e c inode p0 buf r (bsize-r)
    ++ (fun (l,eof,_) ->
	  if not eof && (r+l) < bsize then
	    read_block_e (Int64.add p0 (Int64.of_int l)) buf (r+l) bsize
	  else
	    eps_e (`Done(r+l,eof)) esys
       )
  in

  let rec fetch() =
    if !n < par then (
      if !p < end_pos then (
	let buf = Netsys_mem.pool_alloc_memory pool in
	let p0 = !p in
	let bsize = Int64.to_int (min blocksizeL (Int64.sub end_pos p0)) in
	let _e =
	  read_block_e p0 (`Memory buf) 0 bsize
	  ++ (fun (l,eof) ->
		ignore(Unix.LargeFile.lseek fd p0 Unix.SEEK_SET);
		ignore(Netsys_mem.mem_write fd buf 0 l);
		decr n;
		fetch();
		eps_e (`Done ()) esys
	     ) in
	incr n;
	p := Int64.add !p blocksizeL;
	fetch()
      )
    );
    if !n=0 then signal(`Done())
  in

  fetch();
  let _e =
    stream_e ++ (fun () -> Plasma_client.commit_e t) in
  Unixqueue.run esys


let get() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  let plasma_filename = ref "" in
  let plasma_set = ref false in
  let local_filename = ref "" in
  let local_set = ref false in
  let alt = ref None in

  Arg.parse
    ( ac#args @
	[ "-alt", Arg.Int (fun n -> alt := Some n),
	  "<n>   use the alternate algorithm with a parallelization factor of n"
	]
    )
    (fun s ->
       if not !plasma_set then (
	 plasma_filename := s;
	 plasma_set := true
       )
       else
	 if not !local_set then (
	   local_filename := s;
	   local_set := true
	 )
	 else
	   raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma get <plasma_filename> <local_filename>";

  if not !plasma_set then
    failwith "No plasma filename given";   
  if not !local_set then
    failwith "No local filename given";

  let fd = 
    Unix.openfile 
      !local_filename [Unix.O_WRONLY; Unix.O_TRUNC; Unix.O_CREAT] 0o666 in
  
  let c = ac#open_cluster () in

  let t = Plasma_client.start c in
  let inode = Plasma_client.lookup t !plasma_filename false in
  let ii = Plasma_client.get_inodeinfo t inode in
  Plasma_client.commit t;
  ( match !alt with
      | None ->
	  Plasma_client.copy_out c inode 0L fd ii.eof
      | Some par ->
	  alt_copy_out esys c inode 0L fd ii.eof par
  );
  Unix.close fd;

  eprintf "Copied: %s -> %s\n%!" !plasma_filename !local_filename


let cat() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  let files = ref [] in

  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma get <plasma_filename> <local_filename>";

  let c = ac#open_cluster () in

  List.iter
    (fun file ->
       let t = Plasma_client.start c in
       let inode = Plasma_client.lookup t file false in
       let ii = Plasma_client.get_inodeinfo t inode in
       Plasma_client.commit t;
       Plasma_client.copy_out c inode 0L Unix.stdout ii.eof
    )
    !files


let fsstat() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  Arg.parse
    ac#args
    (fun s ->
       raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma fsstat";

  let c = ac#open_cluster () in
  let st = Plasma_client.fsstat c in
  printf "Total:        %20Ld blocks\n" st.total_blocks;
  printf "Used:         %20Ld blocks\n" st.used_blocks;
  printf "Transitional: %20Ld blocks\n" st.trans_blocks;
  printf "Free:         %20Ld blocks\n"
    (Int64.sub (Int64.sub st.total_blocks st.used_blocks) st.trans_blocks);
  ()


module RSet = 
  Plasma_util.RangeSet(struct
			 type t = int64
			 let compare = Int64.compare
			 let pred n =
			   if n = Int64.min_int then
			     None
			   else
			     Some(Int64.pred n)
			 let succ n =
			   if n = Int64.max_int then
			     None
			   else
			     Some(Int64.succ n)
		       end
		      )


let blocks() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in
  let files = ref [] in

  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma blocks <file> ...";

  let c = ac#open_cluster () in
  let t = Plasma_client.start c in

  List.iter
    (fun file ->
       let inode = Plasma_client.lookup t file false in
       let ii = Plasma_client.get_inodeinfo t inode in
       (* FIXME: very large blocklists *)
       let list = 
	 Plasma_client.get_blocklist t inode 0L (Int64.to_int ii.blocklimit) in
       printf "%s:\n" file;

       (* Analyze. First find ranges *)
       let rset = ref RSet.empty in
       let ht = Hashtbl.create 100 in
       let replication = ref max_int in
       List.iter
	 (fun block ->
	    rset := RSet.add block.index !rset;
	    let l = 
	      try Hashtbl.find ht block.index with Not_found -> [] in
	    Hashtbl.replace ht block.index (block :: l)
	 )
	 list;

       let count_replication i =
	 let l = try Hashtbl.find ht i with Not_found -> [] in
	 List.fold_left
	   (fun acc b -> if not b.node_alive then acc else acc+1)
	   0
	   l
       in

       let dead = -9L in

       let short_info i =
	 let l = try Hashtbl.find ht i with Not_found -> [] in
	 let l' = List.sort (fun b1 b2 -> String.compare b1.node b2.node) l in
	 List.map
	   (fun b ->
	      if b.node_alive 
	      then ( b.node, i, i, b.block, b.block )
	      else ( b.node, i, i, dead, dead )
	   )
	   l'
       in

       let add_info si1 si2 =  (* may raise Not_found if not possible *)
	 if List.length si1 <> List.length si2 then raise Not_found;
	 List.map2
	   (fun ( node1, i1_s, i1_e, block1_s, block1_e )
                ( node2, i2_s, i2_e, block2_s, block2_e ) ->
	      if (node1 = node2 && Int64.succ i1_e = i2_s && 
		  Int64.succ block1_e = block2_s)
	      then
		(node1, i1_s, i2_e, block1_s, block2_e)
	      else
		raise Not_found
	   )
	   si1
	   si2 in

       let print_info si =
	 if si <> [] then (
	   let (_, i_s, i_e, _, _) = List.hd si in
	   printf "%9Ld - %9Ld:" i_s i_e;
	   List.iter
	     (fun (n, _, _, b_s, b_e) ->
		if b_s = dead then
		  printf " %s[DEAD]" n
		else
		  printf " %s[%Ld-%Ld]" n b_s b_e
	     )
	     si;
	   printf "\n";
	 ) in

       (* Iterate over ranges *)
       RSet.iter
	 (fun i1 i2 ->
	    let si_acc = ref [] in
	    let i = ref i1 in
	    while !i <= i2 do
	      let si = short_info !i in
	      ( try
		  si_acc := add_info !si_acc si
		with
		  | Not_found ->
		      print_info !si_acc;
		      si_acc := si
	      );
	      i := Int64.succ !i
	    done;
	    print_info !si_acc
	 )
	 !rset;

       Hashtbl.iter
	 (fun i _ ->
	    let r = count_replication i in
	    replication := min !replication r
	 )
	 ht;

       let r_s =
	 if list = [] then
	   "n/a"
	 else
	   string_of_int !replication in

       printf "  blocks:                %d\n" (List.length list);
       printf "  actual replication:    %s\n" r_s;
       printf "  requested replication: %d\n" ii.replication;
       printf "\n";
    )
    !files;

  Plasma_client.commit t



let help() =
  usage()


let dispatch() =
  if Array.length Sys.argv < 2 then
    failwith "Bad usage";

  let mode_str = Sys.argv.(1) in
  incr Arg.current;

  match mode_str with
    | "create" -> create()
    | "delete" | "rm" -> delete()
    | "mkdir" -> mkdir()
    | "put" -> put()
    | "get" -> get()
    | "cat" -> cat()
    | "list" | "ls" -> list()
    | "link" | "ln" -> link()
    | "rename" | "mv" -> rename()
    | "fsstat" -> fsstat()
    | "blocks" -> blocks()
    | "help" | "-help" | "--help" -> help()
    | _ -> failwith ("Bad usage")

let main () =
  try
    dispatch()
  with
    | Plasma_client.Plasma_error code ->
        prerr_endline ("plasma: " ^ Plasma_util.string_of_errno code);
        exit 2

    | Failure m ->
        prerr_endline ("plasma: " ^ m);
        usage();
        exit 2
          
let () = 
  let ctrl = Gc.get() in
  let ctrl' = { ctrl with 
                  Gc.max_overhead = 1_000_000;  (* Turn compactor off *)
                  Gc.space_overhead = 300;      (* allow more wasted mem *)
                  Gc.minor_heap_size = 65536;   (* bigger minor heap *)
              } in
  Gc.set ctrl';
  main()
